@@ -211,7 +211,7 @@ function _chol!(A::AbstractMatrix, ::Type{UpperTriangular})
             A[k,k] = Akk
             AkkInv = inv(copy(Akk'))
             for j = k + 1:n
-                for i = 1:k - 1
+                @simd for i = 1:k - 1
                     A[k,j] -= A[i,k]'A[i,j]
                 end
                 A[k,j] = AkkInv*A[k,j]
@@ -236,14 +236,15 @@ function _chol!(A::AbstractMatrix, ::Type{LowerTriangular})
                 return LowerTriangular(A), convert(BlasInt, k)
             end
             A[k,k] = Akk
-            AkkInv = inv(Akk)
+            AkkInv = inv(copy(Akk'))
             for j = 1:k - 1
+                Akjc = A[k,j]'
                 @simd for i = k + 1:n
-                    A[i,k] -= A[i,j]*A[k,j]'
+                    A[i,k] -= A[i,j]*Akjc
                 end
             end
-            for i = k + 1:n
-                A[i,k] *= AkkInv'
+            @simd for i = k + 1:n
+                A[i,k] *= AkkInv
             end
         end
      end
@@ -255,16 +256,164 @@ function _chol!(x::Number, _)
     rx = real(x)
     iszero(rx) && return (rx, convert(BlasInt, 1))
     rxr = sqrt(abs(rx))
-    rval =  convert(promote_type(typeof(x), typeof(rxr)), rxr)
+    rval = convert(promote_type(typeof(x), typeof(rxr)), rxr)
     return (rval, convert(BlasInt, rx != abs(x)))
 end
 
-## for StridedMatrices, check that matrix is symmetric/Hermitian
+# _cholpivoted!. Internal methods for calling pivoted Cholesky
+Base.@propagate_inbounds function _swap_rowcols!(A, ::Type{UpperTriangular}, n, j, q)
+    j == q && return
+    @assert j < q
+    # swap rows and cols without touching the possibly undef-ed triangle
+    A[q, q] = A[j, j]
+    for k in 1:j-1 # initial vertical segments
+        A[k,j], A[k,q] = A[k,q], A[k,j]
+    end
+    for k in j+1:q-1 # intermediate segments
+        A[j,k], A[k,q] = conj(A[k,q]), conj(A[j,k])
+    end
+    A[j,q] = conj(A[j,q]) # corner case
+    for k in q+1:n # final horizontal segments
+        A[j,k], A[q,k] = A[q,k], A[j,k]
+    end
+    return
+end
+Base.@propagate_inbounds function _swap_rowcols!(A, ::Type{LowerTriangular}, n, j, q)
+    j == q && return
+    @assert j < q
+    # swap rows and cols without touching the possibly undef-ed triangle
+    A[q, q] = A[j, j]
+    for k in 1:j-1 # initial horizontal segments
+        A[j,k], A[q,k] = A[q,k], A[j,k]
+    end
+    for k in j+1:q-1 # intermediate segments
+        A[k,j], A[q,k] = conj(A[q,k]), conj(A[k,j])
+    end
+    A[q,j] = conj(A[q,j]) # corner case
+    for k in q+1:n # final vertical segments
+        A[k,j], A[k,q] = A[k,q], A[k,j]
+    end
+    return
+end
+### BLAS/LAPACK element types
+_cholpivoted!(A::StridedMatrix{<:BlasFloat}, ::Type{UpperTriangular}, tol::Real, check::Bool) =
+    LAPACK.pstrf!('U', A, tol)
+_cholpivoted!(A::StridedMatrix{<:BlasFloat}, ::Type{LowerTriangular}, tol::Real, check::Bool) =
+    LAPACK.pstrf!('L', A, tol)
+## Non BLAS/LAPACK element types (generic)
+function _cholpivoted!(A::AbstractMatrix, ::Type{UpperTriangular}, tol::Real, check::Bool)
+    rTA = real(eltype(A))
+    # checks
+    Base.require_one_based_indexing(A)
+    n = checksquare(A)
+    # initialization
+    piv = collect(1:n)
+    dots = zeros(rTA, n)
+    temp = similar(dots)
+
+    @inbounds begin
+        # first step
+        Akk, q = findmax(i -> real(A[i,i]), 1:n)
+        stop = tol < 0 ? eps(rTA)*n*abs(Akk) : tol
+        Akk ≤ stop && return A, piv, convert(BlasInt, 0), convert(BlasInt, 1)
+        # swap
+        _swap_rowcols!(A, UpperTriangular, n, 1, q)
+        piv[1], piv[q] = piv[q], piv[1]
+        A[1,1] = Akk = sqrt(Akk)
+        AkkInv = inv(copy(Akk'))
+        @simd for j in 2:n
+            A[1, j] *= AkkInv
+        end
+
+        for k in 2:n
+            @simd for j in k:n
+                dots[j] += abs2(A[k-1, j])
+                temp[j] = real(A[j,j]) - dots[j]
+            end
+            Akk, q = findmax(j -> temp[j], k:n)
+            Akk ≤ stop && return A, piv, convert(BlasInt, k - 1), convert(BlasInt, 1)
+            q += k - 1
+            # swap
+            _swap_rowcols!(A, UpperTriangular, n, k, q)
+            dots[k], dots[q] = dots[q], dots[k]
+            piv[k], piv[q] = piv[q], piv[k]
+            # update
+            A[k,k] = Akk = sqrt(Akk)
+            AkkInv = inv(copy(Akk'))
+            for j in (k+1):n
+                @simd for i in 1:(k-1)
+                    A[k,j] -= A[i,k]'A[i,j]
+                end
+                A[k,j] = AkkInv * A[k,j]
+            end
+        end
+        return A, piv, convert(BlasInt, n), convert(BlasInt, 0)
+    end
+end
+function _cholpivoted!(A::AbstractMatrix, ::Type{LowerTriangular}, tol::Real, check::Bool)
+    rTA = real(eltype(A))
+    # checks
+    Base.require_one_based_indexing(A)
+    n = checksquare(A)
+    # initialization
+    piv = collect(1:n)
+    dots = zeros(rTA, n)
+    temp = similar(dots)
+
+    @inbounds begin
+        # first step
+        Akk, q = findmax(i -> real(A[i,i]), 1:n)
+        stop = tol < 0 ? eps(rTA)*n*abs(Akk) : tol
+        Akk ≤ stop && return A, piv, convert(BlasInt, 0), convert(BlasInt, 1)
+        # swap
+        _swap_rowcols!(A, LowerTriangular, n, 1, q)
+        piv[1], piv[q] = piv[q], piv[1]
+        A[1,1] = Akk = sqrt(Akk)
+        AkkInv = inv(copy(Akk'))
+        @simd for i in 2:n
+            A[i,1] *= AkkInv
+        end
+
+        for k in 2:n
+            @simd for j in k:n
+                dots[j] += abs2(A[j, k-1])
+                temp[j] = real(A[j,j]) - dots[j]
+            end
+            Akk, q = findmax(i -> temp[i], k:n)
+            Akk ≤ stop && return A, piv, convert(BlasInt, k-1), convert(BlasInt, 1)
+            q += k - 1
+            # swap
+            _swap_rowcols!(A, LowerTriangular, n, k, q)
+            dots[k], dots[q] = dots[q], dots[k]
+            piv[k], piv[q] = piv[q], piv[k]
+            # update
+            A[k,k] = Akk = sqrt(Akk)
+            for j in 1:(k-1)
+                Akjc = A[k,j]'
+                @simd for i in (k+1):n
+                    A[i,k] -= A[i,j]*Akjc
+                end
+            end
+            AkkInv = inv(copy(Akk'))
+            @simd for i in (k+1):n
+                A[i, k] *= AkkInv
+            end
+        end
+        return A, piv, convert(BlasInt, n), convert(BlasInt, 0)
+    end
+end
+function _cholpivoted!(x::Number, tol)
+    rx = real(x)
+    iszero(rx) && return (rx, convert(BlasInt, 1))
+    rxr = sqrt(abs(rx))
+    rval = convert(promote_type(typeof(x), typeof(rxr)), rxr)
+    return (rval, convert(BlasInt, !(rx == abs(x) > tol)))
+end
 
 # cholesky!. Destructive methods for computing Cholesky factorization of real symmetric
 # or Hermitian matrix
 ## No pivoting (default)
-function cholesky!(A::RealHermSymComplexHerm, ::NoPivot = NoPivot(); check::Bool = true)
+function cholesky!(A::SelfAdjoint, ::NoPivot = NoPivot(); check::Bool = true)
     C, info = _chol!(A.data, A.uplo == 'U' ? UpperTriangular : LowerTriangular)
     check && checkpositivedefinite(info)
     return Cholesky(C.data, A.uplo, info)
@@ -295,7 +444,7 @@ Stacktrace:
 function cholesky!(A::AbstractMatrix, ::NoPivot = NoPivot(); check::Bool = true)
     checksquare(A)
     if !ishermitian(A) # return with info = -1 if not Hermitian
-        check && checkpositivedefinite(-1)
+        check && checkpositivedefinite(convert(BlasInt, -1))
         return Cholesky(A, 'U', convert(BlasInt, -1))
     else
         return cholesky!(Hermitian(A), NoPivot(); check = check)
@@ -305,23 +454,15 @@ end
 @deprecate cholesky!(A::RealHermSymComplexHerm, ::Val{false}; check::Bool = true) cholesky!(A, NoPivot(); check) false
 
 ## With pivoting
-### BLAS/LAPACK element types
-function cholesky!(A::RealHermSymComplexHerm{<:BlasReal,<:StridedMatrix},
-                   ::RowMaximum; tol = 0.0, check::Bool = true)
-    AA, piv, rank, info = LAPACK.pstrf!(A.uplo, A.data, tol)
-    C = CholeskyPivoted{eltype(AA),typeof(AA),typeof(piv)}(AA, A.uplo, piv, rank, tol, info)
+### Non BLAS/LAPACK element types (generic).
+function cholesky!(A::SelfAdjoint, ::RowMaximum; tol = 0.0, check::Bool = true)
+    AA, piv, rank, info = _cholpivoted!(A.data, A.uplo == 'U' ? UpperTriangular : LowerTriangular, tol, check)
+    C = CholeskyPivoted(AA, A.uplo, piv, rank, tol, info)
     check && chkfullrank(C)
     return C
 end
-@deprecate cholesky!(A::RealHermSymComplexHerm{<:BlasReal,<:StridedMatrix}, ::Val{true}; kwargs...) cholesky!(A, RowMaximum(); kwargs...) false
-
-### Non BLAS/LAPACK element types (generic). Since generic fallback for pivoted Cholesky
-### is not implemented yet we throw an error
-cholesky!(A::RealHermSymComplexHerm{<:Real}, ::RowMaximum; tol = 0.0, check::Bool = true) =
-    throw(ArgumentError("generic pivoted Cholesky factorization is not implemented yet"))
 @deprecate cholesky!(A::RealHermSymComplexHerm{<:Real}, ::Val{true}; kwargs...) cholesky!(A, RowMaximum(); kwargs...) false
 
-### for AbstractMatrix, check that matrix is symmetric/Hermitian
 """
     cholesky!(A::AbstractMatrix, RowMaximum(); tol = 0.0, check = true) -> CholeskyPivoted
 
@@ -333,12 +474,12 @@ e.g. for integer types.
 function cholesky!(A::AbstractMatrix, ::RowMaximum; tol = 0.0, check::Bool = true)
     checksquare(A)
     if !ishermitian(A)
-        C = CholeskyPivoted(A, 'U', Vector{BlasInt}(),convert(BlasInt, 1),
+        C = CholeskyPivoted(A, 'U', Vector{BlasInt}(), convert(BlasInt, 1),
                             tol, convert(BlasInt, -1))
-        check && chkfullrank(C)
+        check && checkpositivedefinite(convert(BlasInt, -1))
         return C
     else
-        return cholesky!(Hermitian(A), RowMaximum(); tol = tol, check = check)
+        return cholesky!(Hermitian(A), RowMaximum(); tol, check)
     end
 end
 @deprecate cholesky!(A::StridedMatrix, ::Val{true}; kwargs...) cholesky!(A, RowMaximum(); kwargs...) false
@@ -410,6 +551,9 @@ end
 # allow packages like SparseArrays.jl to hook into here and redirect to out-of-place `cholesky`
 _cholesky(A::AbstractMatrix, args...; kwargs...) = cholesky!(A, args...; kwargs...)
 
+# allow cholesky of cholesky
+cholesky(A::Cholesky) = A
+
 ## With pivoting
 """
     cholesky(A, RowMaximum(); tol = 0.0, check = true) -> CholeskyPivoted
@@ -427,7 +571,7 @@ The following functions are available for `CholeskyPivoted` objects:
 [`size`](@ref), [`\\`](@ref), [`inv`](@ref), [`det`](@ref), and [`rank`](@ref).
 
 The argument `tol` determines the tolerance for determining the rank.
-For negative values, the tolerance is the machine precision.
+For negative values, the tolerance is equal to `eps()*size(A,1)*maximum(diag(A))`.
 
 If you have a matrix `A` that is slightly non-Hermitian due to roundoff errors in its construction,
 wrap it in `Hermitian(A)` before passing it to `cholesky` in order to treat it as perfectly Hermitian.
@@ -487,11 +631,16 @@ function Cholesky{T}(C::Cholesky) where T
     Cnew = convert(AbstractMatrix{T}, C.factors)
     Cholesky{T, typeof(Cnew)}(Cnew, C.uplo, C.info)
 end
+Cholesky{T,S}(C::Cholesky) where {T,S<:AbstractMatrix} = Cholesky{T,S}(C.factors, C.uplo, C.info)
 Factorization{T}(C::Cholesky{T}) where {T} = C
 Factorization{T}(C::Cholesky) where {T} = Cholesky{T}(C)
 CholeskyPivoted{T}(C::CholeskyPivoted{T}) where {T} = C
 CholeskyPivoted{T}(C::CholeskyPivoted) where {T} =
-    CholeskyPivoted(AbstractMatrix{T}(C.factors),C.uplo,C.piv,C.rank,C.tol,C.info)
+    CholeskyPivoted(AbstractMatrix{T}(C.factors), C.uplo, C.piv, C.rank, C.tol, C.info)
+CholeskyPivoted{T,S}(C::CholeskyPivoted) where {T,S<:AbstractMatrix} =
+    CholeskyPivoted{T,S,typeof(C.piv)}(C.factors, C.uplo, C.piv, C.rank, C.tol, C.info)
+CholeskyPivoted{T,S,P}(C::CholeskyPivoted) where {T,S<:AbstractMatrix,P<:AbstractVector{<:Integer}} =
+    CholeskyPivoted{T,S,P}(C.factors, C.uplo, C.piv, C.rank, C.tol, C.info)
 Factorization{T}(C::CholeskyPivoted{T}) where {T} = C
 Factorization{T}(C::CholeskyPivoted) where {T} = CholeskyPivoted{T}(C)
 
@@ -515,13 +664,28 @@ copy(C::CholeskyPivoted) = CholeskyPivoted(copy(C.factors), C.uplo, C.piv, C.ran
 size(C::Union{Cholesky, CholeskyPivoted}) = size(C.factors)
 size(C::Union{Cholesky, CholeskyPivoted}, d::Integer) = size(C.factors, d)
 
+function _choleskyUfactor(Cfactors, Cuplo)
+    if Cuplo === 'U'
+        return UpperTriangular(Cfactors)
+    else
+        return copy(LowerTriangular(Cfactors)')
+    end
+end
+function _choleskyLfactor(Cfactors, Cuplo)
+    if Cuplo === 'L'
+        return LowerTriangular(Cfactors)
+    else
+        return copy(UpperTriangular(Cfactors)')
+    end
+end
+
 function getproperty(C::Cholesky, d::Symbol)
     Cfactors = getfield(C, :factors)
     Cuplo    = getfield(C, :uplo)
     if d === :U
-        return UpperTriangular(Cuplo === char_uplo(d) ? Cfactors : copy(Cfactors'))
+        _choleskyUfactor(Cfactors, Cuplo)
     elseif d === :L
-        return LowerTriangular(Cuplo === char_uplo(d) ? Cfactors : copy(Cfactors'))
+        _choleskyLfactor(Cfactors, Cuplo)
     elseif d === :UL
         return (Cuplo === 'U' ? UpperTriangular(Cfactors) : LowerTriangular(Cfactors))
     else
@@ -531,13 +695,18 @@ end
 Base.propertynames(F::Cholesky, private::Bool=false) =
     (:U, :L, :UL, (private ? fieldnames(typeof(F)) : ())...)
 
+function Base.:(==)(C1::Cholesky, C2::Cholesky)
+    C1.uplo == C2.uplo || return false
+    C1.uplo == 'L' ? (C1.L == C2.L) : (C1.U == C2.U)
+end
+
 function getproperty(C::CholeskyPivoted{T}, d::Symbol) where {T}
     Cfactors = getfield(C, :factors)
     Cuplo    = getfield(C, :uplo)
     if d === :U
-        return UpperTriangular(sym_uplo(Cuplo) == d ? Cfactors : copy(Cfactors'))
+        _choleskyUfactor(Cfactors, Cuplo)
     elseif d === :L
-        return LowerTriangular(sym_uplo(Cuplo) == d ? Cfactors : copy(Cfactors'))
+        _choleskyLfactor(Cfactors, Cuplo)
     elseif d === :p
         return getfield(C, :piv)
     elseif d === :P
@@ -554,6 +723,11 @@ end
 Base.propertynames(F::CholeskyPivoted, private::Bool=false) =
     (:U, :L, :p, :P, (private ? fieldnames(typeof(F)) : ())...)
 
+function Base.:(==)(C1::CholeskyPivoted, C2::CholeskyPivoted)
+    (C1.uplo == C2.uplo && C1.p == C2.p) || return false
+    C1.uplo == 'L' ? (C1.L == C2.L) : (C1.U == C2.U)
+end
+
 issuccess(C::Union{Cholesky,CholeskyPivoted}) = C.info == 0
 
 adjoint(C::Union{Cholesky,CholeskyPivoted}) = C
@@ -639,7 +813,7 @@ function rdiv!(B::AbstractMatrix, C::Cholesky)
     end
 end
 
-function LinearAlgebra.rdiv!(B::AbstractMatrix, C::CholeskyPivoted)
+function rdiv!(B::AbstractMatrix, C::CholeskyPivoted)
     n = size(C, 2)
     for i in 1:size(B, 1)
         permute!(view(B, i, 1:n), C.piv)
@@ -713,7 +887,7 @@ end
 
 function chkfullrank(C::CholeskyPivoted)
     if C.rank < size(C.factors, 1)
-        throw(RankDeficientException(C.info))
+        throw(RankDeficientException(C.rank))
     end
 end
 
@@ -791,7 +965,7 @@ function lowrankdowndate!(C::Cholesky, v::AbstractVector)
         s = conj(v[i]/Aii)
         s2 = abs2(s)
         if s2 > 1
-            throw(LinearAlgebra.PosDefException(i))
+            throw(PosDefException(i))
         end
         c = sqrt(1 - abs2(s))
 
@@ -835,3 +1009,30 @@ then `CC = cholesky(C.U'C.U - v*v')` but the computation of `CC` only uses
 `O(n^2)` operations.
 """
 lowrankdowndate(C::Cholesky, v::AbstractVector) = lowrankdowndate!(copy(C), copy(v))
+
+function diag(C::Cholesky{T}, k::Int = 0) where {T}
+    N = size(C, 1)
+    absk = abs(k)
+    iabsk = N - absk
+    z = Vector{T}(undef, iabsk)
+    UL = C.factors
+    if C.uplo == 'U'
+        for i in 1:iabsk
+            z[i] = zero(T)
+            for j in 1:min(i, i+absk)
+                z[i] += UL[j, i]'UL[j, i+absk]
+            end
+        end
+    else
+        for i in 1:iabsk
+            z[i] = zero(T)
+            for j in 1:min(i, i+absk)
+                z[i] += UL[i, j]*UL[i+absk, j]'
+            end
+        end
+    end
+    if !(T <: Real) && k < 0
+        z .= adjoint.(z)
+    end
+    return z
+end
